from zoneinfo import ZoneInfo

import pandas as pd
import pytest

from cosimtlk.simulation.storage import ScheduleStore
from tests.conftest import fake_schedule


@pytest.fixture(scope="function")
def db():
    return ScheduleStore()


def test_store_history_without_tz(db):
    history = fake_schedule(
        cutoff_date=[
            pd.Timestamp("2021-01-01 00:00:00"),
            pd.Timestamp("2021-01-01 01:00:00"),
        ],
        columns=["a", "b"],
    )
    db.store_history(schedule=history)

    schedule = db.get_schedule("schedule")
    assert isinstance(schedule, pd.DataFrame)
    assert schedule.timestamp.dt.tz == ZoneInfo("UTC")
    assert schedule.cutoff_date.dt.tz == ZoneInfo("UTC")


def test_store_history_with_tz(db):
    history = fake_schedule(
        cutoff_date=[
            pd.Timestamp("2021-01-01 00:00:00"),
            pd.Timestamp("2021-01-01 01:00:00"),
        ],
        columns=["a", "b"],
        tz=ZoneInfo("Europe/Brussels"),
    )
    db.store_history(schedule=history)

    schedule = db.get_schedule("schedule")
    assert isinstance(schedule, pd.DataFrame)
    assert schedule.timestamp.dt.tz == ZoneInfo("Europe/Brussels")
    assert schedule.cutoff_date.dt.tz == ZoneInfo("Europe/Brussels")


def test_get_set_schedule_without_tz(db):
    cutoff_date = pd.Timestamp("2021-01-01 00:00:00")
    schedule = fake_schedule(cutoff_date)

    db.store_schedule("schedule", schedule, cutoff_date=cutoff_date)

    schedule = db.get_schedule("schedule")
    assert isinstance(schedule, pd.DataFrame)
    assert schedule.timestamp.dt.tz == ZoneInfo("UTC")
    assert schedule.cutoff_date.dt.tz == ZoneInfo("UTC")


def test_get_set_schedule_with_tz(db):
    cutoff_date = pd.Timestamp("2021-01-01 00:00:00", tz=ZoneInfo("Europe/Brussels"))
    schedule = fake_schedule(cutoff_date, tz=ZoneInfo("Europe/Brussels"))

    db.store_schedule("schedule", schedule, cutoff_date=cutoff_date)

    schedule = db.get_schedule("schedule")
    assert isinstance(schedule, pd.DataFrame)
    assert schedule.timestamp.dt.tz == ZoneInfo("Europe/Brussels")
    assert schedule.cutoff_date.dt.tz == ZoneInfo("Europe/Brussels")


def test_get_last_schedule(db):
    schedule_1 = fake_schedule(pd.Timestamp("2021-01-01 00:00:00"))
    schedule_2 = fake_schedule(pd.Timestamp("2021-01-01 03:00:00"))
    schedule_3 = fake_schedule(pd.Timestamp("2021-01-01 06:00:00"))
    schedule_4 = fake_schedule(pd.Timestamp("2021-01-01 09:00:00"))

    db.store_schedule("schedule", schedule_1, cutoff_date=pd.Timestamp("2021-01-01 00:00:00"))
    db.store_schedule("schedule", schedule_2, cutoff_date=pd.Timestamp("2021-01-01 03:00:00"))
    db.store_schedule("schedule", schedule_3, cutoff_date=pd.Timestamp("2021-01-01 06:00:00"))
    db.store_schedule("schedule", schedule_4, cutoff_date=pd.Timestamp("2021-01-01 09:00:00"))

    schedule = db.get_last_schedule("schedule", pd.Timestamp("2021-01-01 06:00:00"))
    assert isinstance(schedule, pd.DataFrame)
    assert (schedule.cutoff_date == pd.Timestamp("2021-01-01 06:00:00+00:00")).all()


def test_get_last_schedule_not_exact_time(db):
    schedule_1 = fake_schedule(pd.Timestamp("2021-01-01 00:00:00"))
    schedule_2 = fake_schedule(pd.Timestamp("2021-01-01 03:00:00"))
    schedule_3 = fake_schedule(pd.Timestamp("2021-01-01 06:00:00"))
    schedule_4 = fake_schedule(pd.Timestamp("2021-01-01 09:00:00"))

    db.store_schedule("schedule", schedule_1, cutoff_date=pd.Timestamp("2021-01-01 00:00:00"))
    db.store_schedule("schedule", schedule_2, cutoff_date=pd.Timestamp("2021-01-01 03:00:00"))
    db.store_schedule("schedule", schedule_3, cutoff_date=pd.Timestamp("2021-01-01 06:00:00"))
    db.store_schedule("schedule", schedule_4, cutoff_date=pd.Timestamp("2021-01-01 09:00:00"))

    schedule = db.get_last_schedule("schedule", pd.Timestamp("2021-01-01 08:55:00"))
    assert isinstance(schedule, pd.DataFrame)
    assert (schedule.cutoff_date == pd.Timestamp("2021-01-01 06:00:00+00:00")).all()


if __name__ == "__main__":
    pytest.main()



from datetime import datetime, timedelta
from zoneinfo import ZoneInfo

import numpy as np
from simpy.util import start_delayed

from cosimtlk.models import DateTimeLike
from cosimtlk.simulation.entities import Entity
from cosimtlk.simulation.environment import Environment
from cosimtlk.simulation.storage import ObservationStore, ScheduleStore, StateStore
from cosimtlk.simulation.utils import ensure_tz


class Simulator:
    def __init__(
        self,
        *,
        initial_time: DateTimeLike,
        entities: list[Entity] | None = None,
    ) -> None:
        """Simulation runner.

        Organizes the entities processes into a discrete event simulation. If multiple entities would run at the same
        time, the order is determined by the order of the entities in the entities list.

        Args:
            initial_time: The initial time of the simulation. Can be either a datetime or a Timestamp.
            entities: A list of entities inside the simulation.
        """
        initial_time = self._parse_datetime(initial_time)
        initial_timestamp = self._dt_to_timestamp(initial_time)
        self.tzinfo: ZoneInfo = initial_time.tzinfo

        # Create simulation environment
        self._environment = Environment(initial_time=initial_timestamp)

        # Initialize data stores
        self._state = StateStore()
        self._db = ObservationStore()
        self._schedules = ScheduleStore()

        # Initialize entities
        self._initialize_entities(entities=entities)

    def _initialize_entities(self, entities: list[Entity] | None = None) -> None:
        processes = []
        for entity in entities or []:
            entity.initialize(self)
            processes.extend(entity.processes)
        process_delays = np.linspace(0.005, 0.995, len(processes))

        for process, delay in zip(processes, process_delays, strict=True):
            start_delayed(self._environment, process(), delay)

    def __repr__(self) -> str:
        return f"<SimulationRunner t={self.current_timestamp}>"

    @property
    def current_timestamp(self) -> int:
        """The current simulation time as a unix timestamp."""
        return int(self._environment.now)

    @property
    def env(self) -> Environment:
        return self._environment

    @property
    def current_datetime(self) -> datetime:
        """The current simulation time as a timezone aware datetime object."""
        return datetime.fromtimestamp(self.current_timestamp, tz=self.tzinfo)

    @property
    def state(self) -> StateStore:
        return self._state

    @property
    def db(self) -> ObservationStore:
        return self._db

    @property
    def schedules(self) -> ScheduleStore:
        return self._schedules

    @staticmethod
    def _parse_datetime(dt: DateTimeLike) -> DateTimeLike:
        return ensure_tz(dt)

    @staticmethod
    def _dt_to_timestamp(dt: DateTimeLike) -> int:
        """Converts a datetime to a unix timestamp."""
        return int(dt.timestamp())

    @staticmethod
    def _td_to_duration(td: timedelta) -> int:
        """Converts a timedelta to a duration in seconds."""
        return int(td.total_seconds())

    def run(self, until: int | datetime, show_progress_bar: bool = True):  # noqa
        """Runs the simulation until the given timestamp.

        Args:
            until: The timestamp until the simulation should run.
            show_progress_bar: Whether to show a progress bar.
        """
        if isinstance(until, datetime):
            if until.tzinfo is not None and until.tzinfo != self.tzinfo:
                msg = f"Until must be in the same timezone as the initial time. {self.tzinfo} != {until.tzinfo}"
                raise ValueError(msg)
            until = ensure_tz(until, default_tz=self.tzinfo)
            until = self._dt_to_timestamp(until)
        else:
            until = int(until)
        self._environment.run(until=until, show_progress_bar=show_progress_bar)

    def run_for(self, duration: int | timedelta, show_progress_bar: bool = True):  # noqa
        """Runs the simulation for the given duration.

        Args:
            duration: The duration for which the simulation should run.
            show_progress_bar: Whether to show a progress bar.
        """
        if isinstance(duration, timedelta):
            duration = self._td_to_duration(duration)
        else:
            duration = int(duration)
        self.run(until=self.current_timestamp + duration, show_progress_bar=show_progress_bar)



from collections.abc import Iterable
from zoneinfo import ZoneInfo

import numpy as np
import pandas as pd
import pytest

from cosimtlk import FMU
from cosimtlk.simulation.utils import UTC, ensure_tz


@pytest.fixture(scope="session")
def local_fmu():
    fmu = FMU("tests/fixtures/fmus/ModSim.Examples.InputTest.fmu")
    return fmu


@pytest.fixture(scope="function")
def local_fmu_instance(local_fmu):
    instance = local_fmu.instantiate(
        start_time=0,
        step_size=1,
        start_values={},
    )
    yield instance
    instance.close()


def fake_data(
    start: pd.Timestamp = pd.Timestamp("2020-01-01"),  # noqa: B008
    freq: str = "1H",
    periods: int = 10,
    tz: ZoneInfo | None = None,
    columns: Iterable[str] = ("a",),
):
    index = pd.date_range(start, freq=freq, periods=periods, tz=tz)
    data = {column: np.random.random(periods) for column in columns}
    df = pd.DataFrame(index=index, data=data)
    return df


def fake_schedule(
    cutoff_date: pd.Timestamp | Iterable[pd.Timestamp],
    freq: str = "1H",
    periods: int = 10,
    tz: ZoneInfo = UTC,
    columns: Iterable[str] = ("a",),
):
    if isinstance(cutoff_date, pd.Timestamp):
        start = ensure_tz(cutoff_date, default_tz=tz)
        schedule = fake_data(start=start, freq=freq, periods=periods, tz=tz, columns=columns)
        return schedule

    schedules = []
    for made_at_ in cutoff_date:
        start = ensure_tz(made_at_, default_tz=tz)
        schedule_ = (
            fake_data(start=start, freq=freq, periods=periods, tz=tz, columns=columns)
            .rename_axis("timestamp")
            .reset_index()
        )
        schedule_["cutoff_date"] = start
        schedules.append(schedule_)

    schedule = pd.concat(schedules, ignore_index=True)
    return schedule



from datetime import datetime

from pydantic import BaseModel

from cosimtlk.models import FMUInputType


class SimulatorModel(BaseModel):
    id: str  # noqa: A003
    fmu: str
    created_at: datetime


class SimulatorCreateModel(BaseModel):
    start_values: dict[str, FMUInputType]
    start_time: int
    step_size: int



from cosimtlk.simulation.entities.base import Entity
from cosimtlk.simulation.entities.fmu import FMUEntity
from cosimtlk.simulation.entities.generic import GenericProcess
from cosimtlk.simulation.entities.input import Input
from cosimtlk.simulation.entities.stateobserver import StateObserver

__all__ = [
    "Entity",
    "FMUEntity",
    "Input",
    "GenericProcess",
    "StateObserver",
]



from __future__ import annotations

import logging
import shutil
from abc import ABCMeta, abstractmethod
from functools import cached_property
from pathlib import Path
from uuid import uuid4

import fmpy
from fmpy.fmi2 import FMU2Slave
from fmpy.model_description import ModelDescription, ScalarVariable

from cosimtlk.client import SimulatorClient
from cosimtlk.models import FMUCausaltyType, FMUInputType

logger = logging.getLogger(__name__)


class FMUInstanceBase(metaclass=ABCMeta):
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __del__(self):
        self.close()

    @property
    @abstractmethod
    def is_initialized(self) -> bool:
        raise NotImplementedError

    def check_is_initialized(self, msg: str = "") -> None:
        """Check if the FMU instance is initialized.

        Args:
            msg (optional): Message to raise if not initialized.
        """
        if not self.is_initialized:
            msg = "FMU instance is not initialized. " + msg
            raise RuntimeError(msg)

    @property
    @abstractmethod
    def step_size(self) -> int | float:
        raise NotImplementedError

    @property
    @abstractmethod
    def current_time(self) -> int | float:
        raise NotImplementedError

    @abstractmethod
    def close(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def reset(
        self,
        *,
        start_time: int | float,
        step_size: int | float,
        start_values: dict[str, FMUInputType],
    ) -> dict[str, FMUInputType]:
        raise NotImplementedError

    @abstractmethod
    def step(self, *, input_values: dict[str, FMUInputType] | None = None) -> dict[str, FMUInputType]:
        raise NotImplementedError

    @abstractmethod
    def advance(self, until: int, *, input_values: dict[str, FMUInputType] | None = None) -> dict[str, FMUInputType]:
        raise NotImplementedError

    @abstractmethod
    def set_inputs(self, values: dict[str, FMUInputType]) -> None:
        raise NotImplementedError

    @abstractmethod
    def read_outputs(self) -> dict[str, FMUInputType]:
        raise NotImplementedError

    @abstractmethod
    def change_parameters(self, parameters: dict[str, FMUInputType]) -> FMUInstanceBase:
        raise NotImplementedError


class FMUInstance(FMUInstanceBase):
    def __init__(
        self,
        fmu: FMU,
        start_time: int | float,
        step_size: int | float,
        start_values: dict[str, FMUInputType],
    ):
        self._fmu = fmu
        self._current_time = start_time
        self._step_size = step_size

        self._unzipdir = fmpy.extract(self._fmu._fmu_path)
        self._instance = FMU2Slave(
            unzipDirectory=self._unzipdir,
            guid=self._fmu.model_description.guid,
            instanceName=str(uuid4()),
            modelIdentifier=self._fmu.model_description.coSimulation.modelIdentifier,
        )
        # Instantiate FMU
        self._initialized = True
        self._terminated = False

        self._instance.instantiate(visible=False, callbacks=None, loggingOn=False)
        self._initialize(start_values=start_values)

        # Input dict
        self._input_map: dict[str, ScalarVariable] = {input_.name: input_ for input_ in self._fmu.inputs}

        # Create maps for faster read of outputs
        self._output_names: dict[str, list[str]] = {
            "Real": [],
            "Integer": [],
            "Boolean": [],
            "String": [],
        }
        self._output_refs: dict[str, list[int]] = {
            "Real": [],
            "Integer": [],
            "Boolean": [],
            "String": [],
        }
        for output in self._fmu.outputs:
            self._output_names[output.type].append(output.name)
            self._output_refs[output.type].append(output.valueReference)

    def _initialize(self, start_values: dict[str, FMUInputType]) -> None:
        self._instance.setupExperiment(startTime=self._current_time)
        self._instance.enterInitializationMode()
        fmpy.simulation.apply_start_values(self._instance, self._fmu.model_description, start_values=start_values or {})
        self._instance.exitInitializationMode()
        self._initialized = True
        self._terminated = False

    def __repr__(self):
        return f"{self.__class__.__name__}(model_name={self._fmu.model_description.modelName})"

    @property
    def is_initialized(self) -> bool:
        return self._initialized

    @property
    def step_size(self) -> int | float:
        """Return the step size of the FMU."""
        return self._step_size

    @property
    def current_time(self) -> int | float:
        """Return the current time of the FMU."""
        return self._current_time

    def _terminate(self):
        if not self._terminated:
            self._instance.terminate()
            self._instance.reset()
            self._terminated = True

    def close(self):
        """Closes the FMU."""
        self._terminate()
        if self.is_initialized:
            try:
                self._instance.freeInstance()
            except Exception as e:
                logger.error("Could not free FMU instance.")
                logger.exception(e)
            shutil.rmtree(self._unzipdir, ignore_errors=True)
        self._initialized = False

    def reset(
        self,
        *,
        start_time: int | float,
        step_size: int | float,
        start_values: dict[str, FMUInputType],
    ) -> FMUInstanceBase:
        """Reset the FMU.

        Args:
            start_time: Start time of the simulation.
            step_size: Step size of the simulation.
            start_values: Start values of the simulation.

        Returns:
            The FMU instance.
        """
        self._terminate()
        self._current_time = start_time
        self._step_size = step_size
        self._initialize(start_values=start_values)
        return self

    def step(self, *, input_values: dict[str, FMUInputType] | None = None) -> dict[str, FMUInputType]:
        """Do a single step of the FMU.

        Args:
            input_values (optional): Input values for the step. Defaults to None.

        Returns:
            The outputs of the FMU.
        """
        self.check_is_initialized(msg="Cannot call step() on an uninitialized fmu.")

        self.set_inputs(input_values or {})
        self._do_step()
        outputs = self.read_outputs()
        return outputs

    def advance(self, until: int, *, input_values: dict[str, FMUInputType] | None = None) -> dict[str, FMUInputType]:
        """Advance the FMU until a given time.

        Args:
            until: Time to advance to.
            input_values (optional): Input values for the advance. Defaults to None.

        Returns:
            The outputs of the FMU.
        """
        self.check_is_initialized(msg="Cannot call advance() on an uninitialized fmu.")

        if until < self._current_time:
            msg = "Cannot advance time to a time in the past."
            raise ValueError(msg)

        self.set_inputs(input_values or {})

        while self._current_time < until:
            self._do_step()

        outputs = self.read_outputs()
        return outputs

    def _do_step(self):
        self._instance.doStep(
            currentCommunicationPoint=self._current_time,
            communicationStepSize=self._step_size,
        )
        self._current_time += self._step_size

    def set_inputs(self, values: dict[str, FMUInputType]) -> None:
        """Sets the inputs of the FMU.

        Args:
            values: Input values to set. Keys are the names of the FMU inputs.
        """
        self.check_is_initialized(msg="Cannot set inputs on an uninitialized FMU.")

        for input_name, input_value in values.items():
            input_ = self._input_map[input_name]
            self._write_variable(input_, input_value)

    def _write_variable(self, variable: ScalarVariable, value: FMUInputType) -> None:
        variable_type = variable.type
        variable_reference = variable.valueReference

        if variable_type == "Real":
            self._instance.setReal([variable_reference], [float(value)])
        elif variable_type == "Integer":
            self._instance.setInteger([variable_reference], [int(value)])
        elif variable_type == "Boolean":
            self._instance.setBoolean([variable_reference], [bool(value)])
        elif variable_type == "String":
            self._instance.setString([variable_reference], [str(value)])
        elif variable_type == "Enumeration":
            self._instance.setInteger([variable_reference], [int(value)])
        else:
            msg = f"Unknown variable type '{variable_type}' for variable '{variable.name}'."
            raise ValueError(msg)

    def _read_variable(self, variable: ScalarVariable) -> FMUInputType:
        variable_type = variable.type
        variable_reference = variable.valueReference

        if variable_type == "Real":
            return self._instance.getReal([variable_reference])[0]
        elif variable_type == "Integer":
            return self._instance.getInteger([variable_reference])[0]
        elif variable_type == "Boolean":
            return self._instance.getBoolean([variable_reference])[0]
        elif variable_type == "String":
            return self._instance.getString([variable_reference])[0]
        elif variable_type == "Enumeration":
            return self._instance.getInteger([variable_reference])[0]
        msg = f"Unknown variable type '{variable_type}' for variable '{variable.name}'."
        raise ValueError(msg)

    def read_outputs(self) -> dict[str, FMUInputType]:
        """Read the outputs of the FMU.

        Returns:
            The outputs of the FMU.
        """
        self.check_is_initialized(msg="Cannot read outputs on an uninitialized FMU.")

        outputs = {"current_time": self._current_time}

        if self._output_refs["Real"]:
            real_outputs = [float(v) for v in self._instance.getReal(self._output_refs["Real"])]
            outputs.update(dict(zip(self._output_names["Real"], real_outputs, strict=True)))

        if self._output_refs["Integer"]:
            integer_outputs = [int(v) for v in self._instance.getInteger(self._output_refs["Integer"])]
            outputs.update(dict(zip(self._output_names["Integer"], integer_outputs, strict=True)))

        if self._output_refs["Boolean"]:
            boolean_outputs = [bool(v) for v in self._instance.getBoolean(self._output_refs["Boolean"])]
            outputs.update(dict(zip(self._output_names["Boolean"], boolean_outputs, strict=True)))

        if self._output_refs["String"]:
            string_outputs = [str(v) for v in self._instance.getString(self._output_refs["String"])]
            outputs.update(dict(zip(self._output_names["String"], string_outputs, strict=True)))
        return outputs

    def change_parameters(self, parameters: dict[str, FMUInputType]) -> FMUInstance:
        """Change the parameters of the FMU.

        Args:
            parameters: Parameters to change. Keys are the names of the FMU parameters.

        Returns:
            The FMU instance.
        """
        start_values = {
            variable.name: self._read_variable(variable) for variable in self._fmu.model_description.modelVariables
        }
        start_values.update(parameters)
        self.reset(start_values=start_values, start_time=self._current_time, step_size=self._step_size)
        return self


class RemoteFMUInstance(FMUInstanceBase):
    def __init__(
        self,
        fmu: RemoteFMU,
        start_time: int | float,
        step_size: int | float,
        start_values: dict[str, FMUInputType] | None = None,
    ):
        self._fmu = fmu
        self._client = fmu._client

        self._current_time = start_time
        self._step_size = step_size

        self._id = self._client.create_simulator(
            path=self._fmu._path,
            start_time=start_time,
            step_size=step_size,
            start_values=start_values,
        ).id
        self._initialized = True

    def __repr__(self):
        return f"{self.__class__.__name__}(id={self._id}, model_name={self._fmu.model_description.modelName})"

    @property
    def id(self) -> str:  # noqa: A003
        """Return the id of the FMU instance."""
        return self._id

    @property
    def is_initialized(self) -> bool:
        return self._initialized

    @property
    def step_size(self) -> int | float:
        """Return the step size of the FMU."""
        return self._step_size

    @property
    def current_time(self) -> int | float:
        """Return the current time of the FMU."""
        return self._current_time

    def close(self) -> None:
        """Closes the FMU."""
        if self._initialized:
            self._client.delete_simulator(self._id)
            self._id = None
        self._initialized = False

    def reset(
        self,
        *,
        start_time: int | float,
        step_size: int | float,
        start_values: dict[str, FMUInputType],
    ) -> RemoteFMUInstance:
        """Reset the FMU.

        Args:
            start_time: Start time of the simulation.
            step_size: Step size of the simulation.
            start_values: Start values of the simulation.

        Returns:
            The FMU instance.
        """
        self._step_size = step_size
        self._current_time = start_time
        self._client.reset(self._id, start_values=start_values, start_time=start_time, step_size=step_size)
        return self

    def step(self, *, input_values: dict[str, FMUInputType] | None = None) -> dict[str, FMUInputType]:
        """Do a single step of the FMU.

        Args:
            input_values (optional): Input values for the step. Defaults to None.

        Returns:
            The outputs of the FMU.
        """
        outputs = self._client.step(self._id, input_values=input_values)
        self._current_time = outputs["current_time"]
        return outputs

    def advance(self, until: int, *, input_values: dict[str, FMUInputType] | None = None) -> dict[str, FMUInputType]:
        """Advance the FMU until a given time.

        Args:
            until: Time to advance to.
            input_values (optional): Input values for the advance. Defaults to None.

        Returns:
            The outputs of the FMU.
        """
        outputs = self._client.advance(self._id, until=until, input_values=input_values)
        self._current_time = outputs["current_time"]
        return outputs

    def set_inputs(self, values: dict[str, FMUInputType]) -> None:
        """Sets the inputs of the FMU.

        Args:
            values: Input values to set. Keys are the names of the FMU inputs.
        """
        self._client.set_inputs(self._id, input_values=values)

    def read_outputs(self) -> dict[str, FMUInputType]:
        """Read the outputs of the FMU.

        Returns:
            The outputs of the FMU.
        """
        return self._client.get_outputs(self._id)

    def change_parameters(self, parameters: dict[str, FMUInputType]) -> RemoteFMUInstance:
        """Change the parameters of the FMU.

        Args:
            parameters: Parameters to change. Keys are the names of the FMU parameters.

        Returns:
            The FMU instance.
        """
        self._client.change_parameters(self._id, parameters=parameters)
        return self


class FMUBase(metaclass=ABCMeta):
    @cached_property
    @abstractmethod
    def model_description(self) -> ModelDescription:
        raise NotImplementedError

    @abstractmethod
    def instantiate(
        self,
        *,
        start_time: int | float,
        step_size: int | float,
        start_values: dict,
        **kwargs,
    ) -> FMUInstanceBase:
        raise NotImplementedError

    @cached_property
    def inputs(self) -> list[ScalarVariable]:
        """Return the inputs of the FMU.

        Returns:
            List of FMU inputs.
        """
        return [
            variable
            for variable in self.model_description.modelVariables
            if variable.causality == FMUCausaltyType.INPUT
        ]

    @cached_property
    def outputs(self) -> list[ScalarVariable]:
        """Return the outputs of the FMU.

        Returns:
            List of FMU outputs.
        """
        return [
            variable
            for variable in self.model_description.modelVariables
            if variable.causality == FMUCausaltyType.OUTPUT
        ]

    @cached_property
    def parameters(self) -> list[ScalarVariable]:
        """Return the parameters of the FMU.

        Returns:
            List of FMU parameters.
        """
        return [
            variable
            for variable in self.model_description.modelVariables
            if variable.causality == FMUCausaltyType.PARAMETER
        ]


class FMU(FMUBase):
    def __init__(self, path: str | Path):
        fmu_path = Path(path).resolve()
        if not fmu_path.exists():
            msg = f"FMU file not found: {fmu_path}"
            raise FileNotFoundError(msg)
        if not fmu_path.suffix == ".fmu":
            msg = f"FMU file must have .fmu extension: {fmu_path}"
            raise ValueError(msg)

        self._fmu_path = str(fmu_path)

    def __repr__(self):
        return f"{self.__class__.__name__}(path={self._fmu_path})"

    @cached_property
    def model_description(self) -> ModelDescription:
        """Return the model description of the FMU."""
        return fmpy.read_model_description(self._fmu_path)

    def recompile(self) -> None:
        """Recompile the FMU for the current platform."""
        fmpy.util.compile_platform_binary(self._fmu_path)
        logger.info(f"Recompiled FMU for platform {fmpy.platform}.")

    def current_platform_is_supported(self) -> bool:
        """Check if the current platform is supported by the FMU."""
        current_platform = fmpy.platform
        supported_platforms = fmpy.supported_platforms(self._fmu_path)
        return current_platform in supported_platforms

    def instantiate(
        self,
        *,
        start_time: int | float,
        step_size: int | float,
        start_values: dict | None,
        **kwargs,
    ) -> FMUInstanceBase:
        """Instantiate the FMU.

        Args:
            start_time: Start time of the simulation.
            step_size: Step size of the simulation.
            start_values (optional): Start values of the simulation. Defaults to None.
            recompile (optional): Recompile the FMU for the current platform if not supported. Defaults to False.

        Returns:
            FMUInstance: The instantiated FMU.
        """
        recompile = kwargs.get("recompile", False)
        if recompile and not self.current_platform_is_supported():
            logger.warning(f"{fmpy.platform} is not supported by this FMU, recompiling...")
            self.recompile()

        return FMUInstance(self, start_time=start_time, step_size=step_size, start_values=start_values)


class RemoteFMU(FMUBase):
    def __init__(self, path: str, *, client: SimulatorClient):
        self._path = path
        self._client = client

    def __repr__(self):
        return f"{self.__class__.__name__}(path={self._path}, client={self._client.base_url})"

    @cached_property
    def model_description(self) -> ModelDescription:
        """Return the model description of the FMU."""
        model_description_dto = self._client.get_fmu_info(self._path)
        model_description = ModelDescription(**model_description_dto)
        # FIXME: cattrs not working here so it is not a full ModelDescription
        # return cattrs.structure(model_description, ModelDescription)
        model_description.modelVariables = [
            ScalarVariable(**variable) for variable in model_description_dto["modelVariables"]
        ]
        return model_description

    def instantiate(
        self,
        *,
        start_time: int | float,
        step_size: int | float,
        start_values: dict,
        **kwargs,  # noqa
    ) -> FMUInstanceBase:
        """Instantiate the FMU.

        Args:
            start_time: Start time of the simulation.
            step_size: Step size of the simulation.
            start_values: Start values of the simulation.

        Returns:
            Instantiated RemoteFMUInstance.
        """
        return RemoteFMUInstance(
            self,
            start_time=start_time,
            step_size=step_size,
            start_values=start_values,
        )



from __future__ import annotations

import logging
from collections.abc import Callable, Generator
from typing import Any, TYPE_CHECKING

from cosimtlk._fmu import FMUBase
from cosimtlk.models import FMUInputType
from cosimtlk.simulation.entities import Entity

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from cosimtlk.simulation import Simulator


class FMUEntity(Entity):
    def __init__(
        self,
        name: str,
        *,
        fmu: FMUBase,
        start_values: dict[str, FMUInputType],
        fmu_step_size: int,
        simulation_step_size: int,
        namespace: str | None = None,
        input_namespace: str = "inputs",
        output_namespace: str = "outputs",
    ):
        super().__init__(name)
        # Simulator inputs
        self.fmu = fmu
        self.fmu_instance = None
        self.start_values = start_values
        self.fmu_step_size = fmu_step_size
        self.simulation_step_size = simulation_step_size

        self.namespace = namespace or self.name
        self.input_namespace = input_namespace
        self.output_namespace = output_namespace

    @property
    def processes(self) -> list[Callable[[], Generator]]:
        return [self.simulation_process]

    def _store_outputs(self, outputs, namespace: str):
        self.context.state.set(**outputs, namespace=namespace)
        logger.debug(f"{self}: t={self.context.current_datetime}, outputs={outputs}")

    def initialize(self, context: Simulator) -> FMUEntity:
        super().initialize(context)

        self.input_namespace = self.context.state.make_namespace(self.namespace, self.input_namespace)
        self.output_namespace = self.context.state.make_namespace(self.namespace, self.output_namespace)

        self.fmu_instance = self.fmu.instantiate(
            start_values=self.start_values,
            step_size=self.fmu_step_size,
            start_time=self.context.current_timestamp,
        )
        return self

    def simulation_process(self):
        self._store_outputs(self.fmu_instance.read_outputs(), namespace=self.output_namespace)
        while True:
            inputs = self.pre_advance()

            # Advance simulation
            outputs = self.fmu_instance.advance(
                self.context.current_timestamp + self.simulation_step_size, input_values=inputs
            )
            time_until_next_step = self.fmu_instance.current_time - self.context.current_timestamp
            yield self.context.env.timeout(time_until_next_step)

            self.post_advance(outputs)

    def pre_advance(self) -> dict[str, Any]:
        # Collect inputs
        inputs = self.context.state.get_all(namespace=self.input_namespace)
        logger.debug(f"{self}: t={self.context.current_datetime}, inputs={inputs}")
        return inputs

    def post_advance(self, outputs: dict[str, Any]) -> None:
        # Store outputs
        self._store_outputs(outputs, namespace=self.output_namespace)



from datetime import datetime
from enum import Enum

from pandas import Timestamp

FMUInputType = float | int | str | bool
DateTimeLike = datetime | Timestamp


class FMUCausaltyType(str, Enum):
    INPUT = "input"
    OUTPUT = "output"
    PARAMETER = "parameter"
    CALCULATED_PARAMETER = "calculatedParameter"
    LOCAL = "local"


class FMUState(str, Enum):
    STUFF = "stuff"



import logging
from collections.abc import Callable, Generator

from pandas import DataFrame, Series

from cosimtlk.simulation.entities import Entity

logger = logging.getLogger(__name__)


class Input(Entity):
    def __init__(self, name: str, *, data: Series | DataFrame):
        """An entity that sets the input of the simulator based on the input date.

        Args:
            name: The name of the entity.
            data: The input data. Can be a Series or a DataFrame with a DatetimeIndex.
                The index of the data is used as the time at which the input is set,
                while the name of the columns are used as the name of the inputs.

        """
        super().__init__(name)
        if isinstance(data, Series):
            data = data.copy().to_frame()
        self.data = data
        self._index = 0

    @property
    def processes(self) -> list[Callable[[], Generator]]:
        return [self.set_inputs_process]

    def set_inputs_process(self):
        while True:
            if self.data.empty or self._index >= len(self.data):
                logger.warning(f"{self}: t={self.context.current_datetime}, no data left.")
                break

            current_time = self.context.current_datetime
            next_point_at = self.data.index[self._index]

            if next_point_at <= current_time:
                next_points = self.data.iloc[self._index].to_dict()
                logger.debug(f"{self}: t={self.context.current_datetime}, setting inputs: {next_points}")
                self.context.state.set(**next_points)
                self._index += 1
            else:
                next_point_in = int((next_point_at - current_time).total_seconds())
                yield self.context.env.timeout(next_point_in)



from __future__ import annotations

import functools
from collections.abc import Callable, Generator

from cosimtlk.simulation.entities import Entity


class StateObserver(Entity):
    def __init__(
        self,
        name: str,
        *,
        measurements: list[str],
        scheduler: Callable,
    ):
        """An entity that stores the current state of the simulation into long term storage.

        Args:
            name: The name of the entity.
            measurements: A mapping of state names to measurement names. The keys are used as
                the names of the measurements inside the database, while the values determine the state.
            scheduler: A generator function that schedules a function such as `cosimtlk.simulation.utils.every`
                or `cosimtlk.simulation.utils.cron`.
        """
        super().__init__(name)
        self.measurements = measurements
        self.scheduler = scheduler

    @property
    def processes(self) -> list[Callable[[], Generator]]:
        scheduled_process = self.scheduler(self.__class__.sensing_process)
        return [
            functools.partial(scheduled_process, self),
        ]

    def sensing_process(self):
        values = {measurement: self.context.state.get(measurement) for measurement in self.measurements}
        timestamp = self.context.current_datetime
        self.context.db.store_observations(timestamp, **values)



from cosimtlk._fmu import FMU, FMUInstance, RemoteFMU, RemoteFMUInstance
from cosimtlk.client import SimulatorClient

__all__ = [
    "FMU",
    "FMUInstance",
    "RemoteFMU",
    "RemoteFMUInstance",
    "SimulatorClient",
]
